
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program plaq_action().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "su3fcts.h"
#include "aflds.h"
#include "global.h"

#define N0 (NPROC0*L0)

static int bc,nfc[8],ofs[8];
static su3_alg_dble *g,*gbuf;
static su3_dble vd ALIGNED16;
static su3_dble X ALIGNED16;
static su3_dble Y ALIGNED16;


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));
}


static void pack_gbuf(void)
{
   int ifc,ib,ix;

   nfc[0]=FACE0/2;
   nfc[1]=FACE0/2;
   nfc[2]=FACE1/2;
   nfc[3]=FACE1/2;
   nfc[4]=FACE2/2;
   nfc[5]=FACE2/2;
   nfc[6]=FACE3/2;
   nfc[7]=FACE3/2;

   ofs[0]=0;
   ofs[1]=ofs[0]+nfc[0];
   ofs[2]=ofs[1]+nfc[1];
   ofs[3]=ofs[2]+nfc[2];
   ofs[4]=ofs[3]+nfc[3];
   ofs[5]=ofs[4]+nfc[4];
   ofs[6]=ofs[5]+nfc[5];
   ofs[7]=ofs[6]+nfc[6];

   for (ifc=0;ifc<8;ifc++)
   {
      for (ib=0;ib<nfc[ifc];ib++)
      {
         ix=map[ofs[ifc]+ib];
         gbuf[ofs[ifc]+ib]=g[ix];
      }
   }
}


static void send_gbuf(void)
{
   int ifc,np,saddr,raddr;
   int nbf,tag;
   su3_alg_dble *sbuf,*rbuf;
   MPI_Status stat;

   np=cpr[0]+cpr[1]+cpr[2]+cpr[3];

   for (ifc=0;ifc<8;ifc++)
   {
      nbf=8*nfc[ifc];

      if (nbf>0)
      {
         tag=mpi_tag();
         saddr=npr[ifc^0x1];
         raddr=npr[ifc];
         sbuf=gbuf+ofs[ifc];
         rbuf=g+VOLUME+ofs[ifc];

         if (np&0x1)
         {
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
         }
         else
         {
            MPI_Recv(rbuf,nbf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
            MPI_Send(sbuf,nbf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
         }
      }
   }
}


static void random_g(void)
{
   int ix,t;
   su3_alg_dble *gx;

   gauss_dble((double*)(g+NSPIN),16);
   MPI_Bcast(g+NSPIN,16,MPI_DOUBLE,0,MPI_COMM_WORLD);

   gx=g;

   for (ix=0;ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if ((t>0)||(bc!=1))
         gauss_dble((double*)(gx),8);
      else
         (*gx)=g[NSPIN];

      gx+=1;
   }

   if (BNDRY>0)
   {
      pack_gbuf();
      send_gbuf();
   }
}


static void transform(su3_alg_dble *ad)
{
   int ix,iy,t,ifc;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      if (t==0)
      {
         iy=iup[ix][0];
         _su3_alg_sub_assign((*ad),g[iy]);
         _su3_alg_add_assign((*ad),g[ix]);
         ad+=1;

         iy=idn[ix][0];
         _su3_alg_sub_assign((*ad),g[NSPIN+1]);
         _su3_alg_add_assign((*ad),g[iy]);
         ad+=1;

         for (ifc=2;ifc<8;ifc++)
         {
            if (bc!=1)
            {
               if (ifc&0x1)
               {
                  iy=idn[ix][ifc/2];
                  _su3_alg_sub_assign((*ad),g[ix]);
                  _su3_alg_add_assign((*ad),g[iy]);
               }
               else
               {
                  iy=iup[ix][ifc/2];
                  _su3_alg_sub_assign((*ad),g[iy]);
                  _su3_alg_add_assign((*ad),g[ix]);
               }
            }

            ad+=1;
         }
      }
      else if (t==(N0-1))
      {
         _su3_alg_sub_assign((*ad),g[NSPIN+1]);
         _su3_alg_add_assign((*ad),g[ix]);
         ad+=1;

         for (ifc=1;ifc<8;ifc++)
         {
            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];
               _su3_alg_sub_assign((*ad),g[ix]);
               _su3_alg_add_assign((*ad),g[iy]);
            }
            else
            {
               iy=iup[ix][ifc/2];
               _su3_alg_sub_assign((*ad),g[iy]);
               _su3_alg_add_assign((*ad),g[ix]);
            }

            ad+=1;
         }
      }
      else
      {
         for (ifc=0;ifc<8;ifc++)
         {
            if (ifc&0x1)
            {
               iy=idn[ix][ifc/2];
               _su3_alg_sub_assign((*ad),g[ix]);
               _su3_alg_add_assign((*ad),g[iy]);
            }
            else
            {
               iy=iup[ix][ifc/2];
               _su3_alg_sub_assign((*ad),g[iy]);
               _su3_alg_add_assign((*ad),g[ix]);
            }

            ad+=1;
         }
      }
   }
}


static void global_transform(su3_alg_dble *ad)
{
   int ix,ifc;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         su3alg_to_su3(ad,&X);
         su3xsu3(&vd,&X,&Y);
         su3xsu3dag(&Y,&vd,&X);
         su3_to_su3alg(&X,ad);
         ad+=1;
      }
   }
}


static void random_vec(int *svec)
{
   int mu,bs[4];
   double r[4];

   bs[0]=NPROC0*L0;
   bs[1]=NPROC1*L1;
   bs[2]=NPROC2*L2;
   bs[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(bs[mu])*r[mu]);
         if (svec[mu]>(bs[mu]/2))
            svec[mu]-=bs[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int my_rank,n,s[4];
   double p1,p2;
   su3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);

      printf("\n");
      printf("Check of the program plaq_action()\n");
      printf("----------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check1.c]",
                    "Syntax: check1 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();
   alloc_wad(2);
   alloc_wcd(2);

   g=amalloc((NSPIN+2)*sizeof(*g),ALIGN);
   if (BNDRY>0)
      gbuf=amalloc((BNDRY/2)*sizeof(*gbuf),ALIGN);
   else
      gbuf=NULL;
   error((g==NULL)||((BNDRY>0)&&(gbuf==NULL)),1,"main [check1.c]",
         "Unable to allocate auxiliary arrays");

   wad=reserve_wad(2);
   p1=plaq_action(wad[0],wad[1]);

   if (my_rank==0)
      printf("plaq_action after field initialization: %.1e\n",p1);

   new_fld(wad[0]);
   new_fld(wad[1]);
   set_adbc(0,wad[0]);
   set_adbc(0,wad[1]);
   p1=plaq_action(wad[0],wad[1]);

   if (my_rank==0)
      printf("plaq_action of random field: %.15e\n\n",p1);

   random_g();
   transform(wad[0]);
   transform(wad[1]);
   p2=plaq_action(wad[0],wad[1]);

   if (my_rank==0)
   {
      p2-=p1;
      printf("Local gauge invariance:\n");
      printf("Relative difference (plaq_action) = %.1e\n\n",fabs(p2/p1));
   }

   random_su3_dble(&vd);
   MPI_Bcast(&vd,18,MPI_DOUBLE,0,MPI_COMM_WORLD);
   new_fld(wad[0]);
   new_fld(wad[1]);
   set_adbc(0,wad[0]);
   set_adbc(0,wad[1]);
   p1=plaq_action(wad[0],wad[1]);
   global_transform(wad[0]);
   global_transform(wad[1]);
   p2=plaq_action(wad[0],wad[1]);

   if (my_rank==0)
   {
      p2-=p1;
      printf("Global gauge invariance:\n");
      printf("Relative difference (plaq_action) = %.1e\n\n",fabs(p2/p1));
   }

   if (my_rank==0)
      printf("Translation invariance:\n");

   new_fld(wad[0]);
   new_fld(wad[1]);
   set_adbc(0,wad[0]);
   set_adbc(0,wad[1]);
   p1=plaq_action(wad[0],wad[1]);

   for (n=0;n<4;n++)
   {
      random_vec(s);
      shift_ad(s,wad[0]);
      shift_ad(s,wad[1]);
      p2=plaq_action(wad[0],wad[1]);

      if (my_rank==0)
      {
         printf("s=(% 3d,% 3d,% 3d,% 3d):\n",s[0],s[1],s[2],s[3]);
         p2-=p1;
         printf("Relative difference (plaq_action) = %.1e\n",fabs(p2/p1));
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
