
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs gather_afld() and scatter_afld().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "aflds.h"
#include "global.h"


static void random_flds(int n,su3_alg_dble **wad)
{
   int i;

   for (i=0;i<n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(wad[i]),8*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(wad[i]),8*(4*VOLUME+7*(BNDRY/4)));
   }
}


static void set_lks(su3_alg_dble *ad)
{
   int x0,x1,x2,x3,ix,mu;
   int n[4],bo[4],x[4],r;
   su3_alg_dble *X;

   n[0]=NPROC0*L0;
   n[1]=NPROC1*L1;
   n[2]=NPROC2*L2;
   n[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  x[0]=x0;
                  x[1]=x1;
                  x[2]=x2;
                  x[3]=x3;

                  X=ad+8*(ix-(VOLUME/2));

                  for (mu=0;mu<4;mu++)
                  {
                     (*X).c1=(double)(mu);
                     (*X).c2=(double)(bo[0]+x[0]);
                     (*X).c3=(double)(bo[1]+x[1]);
                     (*X).c4=(double)(bo[2]+x[2]);
                     (*X).c5=(double)(bo[3]+x[3]);
                     X+=1;

                     r=x[mu];
                     x[mu]-=1;

                     if ((bo[mu]==0)&&(x[mu]<0))
                        x[mu]+=n[mu];

                     (*X).c1=(double)(mu);
                     (*X).c2=(double)(bo[0]+x[0]);
                     (*X).c3=(double)(bo[1]+x[1]);
                     (*X).c4=(double)(bo[2]+x[2]);
                     (*X).c5=(double)(bo[3]+x[3]);
                     X+=1;
                     x[mu]=r;
                  }
               }
            }
         }
      }
   }
}


static void check_gather(su3_alg_dble **sad)
{
   int mu,ix,iy,x0,x1,x2,x3;
   int bo[4],ie;
   su3_alg_dble *X;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (mu=0;mu<4;mu++)
   {
      X=sad[mu];

      for (ix=0;ix<VOLUME;ix++)
      {
         iy=ix;
         x3=iy%L3;
         iy/=L3;
         x2=iy%L2;
         iy/=L2;
         x1=iy%L1;
         iy/=L1;
         x0=iy;

         ie|=((*X).c1!=(double)(mu));
         ie|=((*X).c2!=(double)(bo[0]+x0));
         ie|=((*X).c3!=(double)(bo[1]+x1));
         ie|=((*X).c4!=(double)(bo[2]+x2));
         ie|=((*X).c5!=(double)(bo[3]+x3));

         X+=1;
      }

      error(ie!=0,1,"check_gather [check2.c]",
            "Field variables are not correctly copied");
   }
}


static void check_scatter(su3_alg_dble *ad)
{
   int x0,x1,x2,x3,ix,mu;
   int n[4],bo[4],x[4],r,ie;
   su3_alg_dble *X;

   n[0]=NPROC0*L0;
   n[1]=NPROC1*L1;
   n[2]=NPROC2*L2;
   n[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  x[0]=x0;
                  x[1]=x1;
                  x[2]=x2;
                  x[3]=x3;

                  X=ad+8*(ix-(VOLUME/2));

                  for (mu=0;mu<4;mu++)
                  {
                     ie|=((*X).c1!=(double)(mu));
                     ie|=((*X).c2!=(double)(bo[0]+x[0]));
                     ie|=((*X).c3!=(double)(bo[1]+x[1]));
                     ie|=((*X).c4!=(double)(bo[2]+x[2]));
                     ie|=((*X).c5!=(double)(bo[3]+x[3]));
                     X+=1;

                     r=x[mu];
                     x[mu]-=1;

                     if ((bo[mu]==0)&&(x[mu]<0))
                        x[mu]+=n[mu];

                     ie|=((*X).c1!=(double)(mu));
                     ie|=((*X).c2!=(double)(bo[0]+x[0]));
                     ie|=((*X).c3!=(double)(bo[1]+x[1]));
                     ie|=((*X).c4!=(double)(bo[2]+x[2]));
                     ie|=((*X).c5!=(double)(bo[3]+x[3]));
                     X+=1;
                     x[mu]=r;
                  }
               }
            }
         }
      }
   }

   error(ie!=0,1,"check_scatter [check2.c]",
         "Field variables are not correctly copied");
}



int main(int argc,char *argv[])
{
   int my_rank;
   su3_alg_dble **wad,*sad[4];
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);

      printf("\n");
      printf("Check of the programs gather_afld() and scatter_afld()\n");
      printf("------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wad(3);
   wad=reserve_wad(3);
   random_flds(3,wad);

   set_lks(wad[0]);

   sad[0]=wad[1];
   sad[1]=sad[0]+VOLUME;
   sad[2]=sad[1]+VOLUME;
   sad[3]=sad[2]+VOLUME;

   gather_afld(wad[0],sad);
   check_gather(sad);

   scatter_afld(sad,wad[2]);
   check_scatter(wad[2]);

   release_wad();

   if (my_rank==0)
   {
      printf("No errors discovered\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
