
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs aprop() and random_asrc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "linalg.h"
#include "aflds.h"
#include "global.h"

static int *rlxs_state=NULL,*rlxd_state;
static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];


static void save_ranlux(void)
{
   int nlxs,nlxd;

   if (rlxs_state==NULL)
   {
      nlxs=rlxs_size();
      nlxd=rlxd_size();

      rlxs_state=malloc((nlxs+nlxd)*sizeof(int));
      rlxd_state=rlxs_state+nlxs;

      error(rlxs_state==NULL,1,"save_ranlux [check4.c]",
            "Unable to allocate state arrays");
   }

   rlxs_get(rlxs_state);
   rlxd_get(rlxd_state);
}


static void restore_ranlux(void)
{
   rlxs_reset(rlxs_state);
   rlxd_reset(rlxd_state);
}


static void new_fld(su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)));
}


static int is_equal(su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int it;

   it=1;

   it&=((*ad1).c1==(*ad2).c1);
   it&=((*ad1).c2==(*ad2).c2);
   it&=((*ad1).c3==(*ad2).c3);
   it&=((*ad1).c4==(*ad2).c4);
   it&=((*ad1).c5==(*ad2).c5);
   it&=((*ad1).c6==(*ad2).c6);
   it&=((*ad1).c7==(*ad2).c7);
   it&=((*ad1).c8==(*ad2).c8);

   return it;
}


static int cmp_flds(su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int ix,ifc,it;

   it=1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         it&=is_equal(ad1,ad2);
         ad1+=1;
         ad2+=1;
      }
   }

   return it;
}


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check4.c]",
         "Unable to allocate auxiliary array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,order,ie;
   double p1,p2,p3,lambda0;
   su3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);

      printf("\n");
      printf("Check of the programs aprop() and random_asrc()\n");
      printf("-----------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check4.c]",
                    "Syntax: check4 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   order=1;
   lambda0=0.4567;
   set_th_parms(LYM,bc,0);
   default_pv(order);
   pv[(int)(LAMBDA)][0]=lambda0;
   set_nspt_parms(KSPT,order,opv,pv);
   print_th_parms();
   print_nspt_parms(0xffff);
   start_ranlux(0,12345);
   geometry();
   alloc_wad(4);
   alloc_wcd(2);

   wad=reserve_wad(4);
   new_fld(wad[0]);
   new_fld(wad[1]);
   new_fld(wad[2]);
   new_fld(wad[3]);
   set_adbc(0,wad[0]);
   set_adbc(0,wad[1]);

   p1=0.5*scalar_prod_alg(4*VOLUME,1,wad[0],wad[1]);
   assign_alg2alg(4*VOLUME,wad[1],wad[3]);
   aprop(wad[1],wad[2]);

   ie=cmp_flds(wad[1],wad[3]);
   error(ie!=1,1,"main [check4.c]",
         "Input field is not unchanged");

   ie=check_adbc(0,wad[2]);
   error(ie!=1,1,"main [check4.c]",
   "Output of aprop does not satisfy the boundary conditions");

   aprop(wad[1],wad[1]);
   ie=cmp_flds(wad[1],wad[2]);
   error(ie!=1,1,"main [check4.c]",
         "Incorrect in place application of aprop()");

   p2=plaq_action(wad[0],wad[2]);
   p3=lambda0*gfix_action(wad[0],wad[2]);

   if (my_rank==0)
   {
      printf("Check of aprop():\n");
      printf("plaq_action = % .15e\n",p2);
      printf("gfix_action = % .15e\n",p3);
      printf("Relative deviation from action bilinear form = %.1e\n\n",
             fabs((p1-p2-p3)/p1));
   }

   new_fld(wad[0]);
   new_fld(wad[1]);
   assign_alg2alg(4*VOLUME,wad[0],wad[1]);
   set_adbc(0,wad[0]);
   aprop(wad[0],wad[2]);
   aprop(wad[1],wad[3]);
   ie=cmp_flds(wad[2],wad[3]);
   error(ie!=1,1,"main [check4.c]",
         "aprop() is sensitive to the boundary values");

   save_ranlux();
   random_asrc(wad[0]);
   random_asrc(wad[1]);
   ie=check_adbc(0,wad[0]);
   ie&=check_adbc(0,wad[1]);
   error(ie!=1,1,"main [check4.c]",
   "Output of random_asrc does not satisfy the boundary conditions");

   restore_ranlux();
   random_alg(4*VOLUME,wad[2]);
   random_alg(4*VOLUME,wad[3]);
   set_adbc(0,wad[2]);
   set_adbc(0,wad[3]);

   p1=0.5*scalar_prod_alg(4*VOLUME,1,wad[2],wad[3]);
   p2=plaq_action(wad[0],wad[1]);
   p3=lambda0*gfix_action(wad[0],wad[1]);

   if (my_rank==0)
   {
      printf("Check of random_asrc()\n");
      printf("plaq_action = % .15e\n",p2);
      printf("gfix_action = % .15e\n",p3);
      printf("Relative deviation from action bilinear form = %.1e\n\n",
             fabs((p1-p2-p3)/p1));
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
