
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program shift_cd().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "random.h"
#include "lattice.h"
#include "cflds.h"
#include "global.h"

static const su3_alg_dble cd0={0.0,0.0,0.0,0.0,0.0,0.0,0.0,0.0};


static void set_fld(su3_alg_dble *cd)
{
   int x0,x1,x2,x3,ix;
   int bo[4];
   su3_alg_dble *X;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];
               X=cd+ix;

               (*X)=cd0;
               (*X).c1=(double)(bo[0]+x0);
               (*X).c2=(double)(bo[1]+x1);
               (*X).c3=(double)(bo[2]+x2);
               (*X).c4=(double)(bo[3]+x3);
            }
         }
      }
   }
}


static int check_fld(int *svec,su3_alg_dble *cd)
{
   int x0,x1,x2,x3;
   int ix,mu,ie,ieall;
   int N[4],bo[4],x[4],y[4];
   su3_alg_dble *X;

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;
   ie=0;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               x[0]=x0+bo[0];
               x[1]=x1+bo[1];
               x[2]=x2+bo[2];
               x[3]=x3+bo[3];

               for (mu=0;mu<4;mu++)
               {
                  y[mu]=x[mu]-svec[mu];
                  y[mu]=safe_mod(y[mu],N[mu]);
               }

               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];
               X=cd+ix;
               ie|=((*X).c1!=(double)(y[0]));
               ie|=((*X).c2!=(double)(y[1]));
               ie|=((*X).c3!=(double)(y[2]));
               ie|=((*X).c4!=(double)(y[3]));
            }
         }
      }
   }

   MPI_Reduce(&ie,&ieall,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&ieall,1,MPI_INT,0,MPI_COMM_WORLD);

   return ieall;
}


static void random_vec(int *svec)
{
   int mu,N[4];
   double r[4];

   N[0]=NPROC0*L0;
   N[1]=NPROC1*L1;
   N[2]=NPROC2*L2;
   N[3]=NPROC3*L3;
   svec[0]=0;
   svec[1]=0;
   svec[2]=0;
   svec[3]=0;

   while ((abs(svec[1])+abs(svec[2])+abs(svec[3]))==0)
   {
      ranlxd(r,4);

      for (mu=1;mu<4;mu++)
      {
         svec[mu]=(int)((double)(N[mu])*r[mu]);
         if (svec[mu]>(N[mu]/2))
            svec[mu]-=N[mu];
      }
   }

   MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
}


int main(int argc,char *argv[])
{
   int my_rank,ie;
   int n,i,s[4];
   su3_alg_dble **wcd;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);

      printf("\n");
      printf("Check of the program shift_cd()\n");
      printf("-------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   start_ranlux(0,12345);
   geometry();
   alloc_wcd(1);
   wcd=reserve_wcd(1);

   for (i=0;i<8;i++)
   {
      set_fld(wcd[0]);
      random_vec(s);
      n=shift_cd(s,wcd[0]);
      ie=check_fld(s,wcd[0]);

      if (my_rank==0)
      {
         printf("s=(% 3d,% 3d,% 3d,% 3d), n=%2d ",s[0],s[1],s[2],s[3],n);

         if (ie)
            printf("(failed)\n");
         else
            printf("(ok)\n");
      }
   }

   release_wcd();

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
