
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparison of force0() with action0().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3fcts.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "aflds.h"
#include "uflds.h"
#include "linalg.h"
#include "mdint.h"
#include "forces.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define NMAX 4

static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static double dS[NMAX+1],s0[NMAX+1],s1[NMAX+1],sw[NMAX+1];
static su3_alg_dble *Aw;
static su3_dble *Uw;


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check1.c]",
         "Unable to allocated parameter array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CG)]=i;
      pv[(int)(CG)][i]=0.126-0.378*(double)(i*i/(order*order));
   }

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CGP)]=i;
      pv[(int)(CGP)][i]=0.0523+0.117*(double)(i*i/(order*order));
   }

   set_nspt_parms(KSPT,order,opv,pv);
}


static void alloc_arrays(void)
{
   int m;

   m=NMAX+1;
   Aw=amalloc(m*sizeof(*Aw),4);
   Uw=amalloc(m*sizeof(*Uw),4);

   error((Aw==NULL)||(Uw==NULL),1,"alloc_arrays [check1.c]",
         "Unable to allocate auxiliary arrays");
}


static void new_fld(int n,su3_alg_dble **ad)
{
   int i;

   for (i=0;i<=n;i++)
   {
      if (cpr[0]==(NPROC0-1))
         gauss_dble((double*)(ad[i]),8*(4*VOLUME+7*(BNDRY/4)+3));
      else
         gauss_dble((double*)(ad[i]),8*(4*VOLUME+7*(BNDRY/4)));

      set_adbc(0,ad[i]);
   }
}


static void copy_ud(int n,su3_dble *ud,su3_dble *sv)
{
   int ix;

   for (ix=0;ix<(4*VOLUME*(n+1));ix++)
      sv[ix]=ud[ix];
}


static double lin_check(int n,su3_dble *ud)
{
   int m;
   double c,dev;
   su3_alg_dble **wad,*mom,*frc;

   m=n+1;
   wad=reserve_wad(2*m);
   mom=wad[0];
   frc=wad[m];

   c=-1.2345;
   random_mom(n,mom);
   assign_alg2alg(4*VOLUME*m,mom,frc);

   force0(n,ud,c,mom);
   sub_alg(4*VOLUME*m,frc,mom);
   set_alg2zero(4*VOLUME*m,frc);
   force0(n,ud,1.0,frc);
   mulr_add_alg(4*VOLUME*m,-c,frc,mom);

   dev=norm_square_alg(4*VOLUME*m,1,mom)/(4*VOLUME*m);
   release_wad();

   return sqrt(dev);
}


static void rot_ud(int n,double eps,su3_alg_dble **ad,su3_dble *ud)
{
   int bc,m,ix,t,ifc,i;

   bc=bc_type();
   m=n+1;

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for(i=0;i<=n;i++)
               su3alg_mulr(&eps,ad[i]+8*(ix-VOLUME/2)+ifc,Aw+i);

            pexp_su3alg(m,Aw,Uw);
            cm3x3_varmul_poly(m,Uw,ud,ud);
         }

         ud+=m;
      }
   }
}


static void dSdt(int n,su3_alg_dble **ad,su3_dble *ud)
{
   int bc,m,ix,t,ifc;
   int i,j,ism;
   su3_alg_dble **wad,*frc;

   bc=bc_type();
   m=n+1;

   wad=reserve_wad(m);
   frc=wad[0];
   set_alg2zero(4*VOLUME*m,frc);
   force0(n,ud,1.0,frc);

   ism=init_hsum(m);
   reset_hsum(ism);

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
         {
            for(i=0;i<=n;i++)
               Aw[i]=ad[i][8*(ix-VOLUME/2)+ifc];

            for (i=0;i<=n;i++)
            {
               dS[i]=0.0;

               for (j=0;j<=i;j++)
                  dS[i]+=su3alg_biform(Aw+j,frc+i-j);

               if ((t==0)&&(ifc>=2))
                  dS[i]*=0.5;
            }

            add_to_hsum(ism,dS);
         }

         frc+=m;
      }
   }

   local_hsum(ism,dS);
   release_wad();
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,i;
   double eps,d,dmax,dev;
   su3_alg_dble **wad;
   su3_dble **wud,**wsv;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);

      printf("\n");
      printf("Comparison of force0() with action0()\n");
      printf("-------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check1.c]",
                    "Syntax: check1 [-bc <type>]");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   default_pv(NMAX);
   print_th_parms();
   print_nspt_parms((1<<(int)(LAMBDA))^0xffff);

   start_ranlux(0,12345);
   geometry();
   alloc_arrays();
   alloc_wad(3*(NMAX+1));
   alloc_wud(2*(NMAX+1));
   wad=reserve_wad(NMAX+1);
   wud=reserve_wud(NMAX+1);
   wsv=reserve_wud(NMAX+1);

   for (n=0;n<=NMAX;n++)
   {
      new_fld(n,wad);
      a2uflds(n,wad,wud[0]);
      copy_ud(n,wud[0],wsv[0]);
      dev=lin_check(n,wud[0]);
      new_fld(n,wad);
      dSdt(n,wad,wud[0]);

      eps=3.0e-4;
      rot_ud(n,eps,wad,wud[0]);
      action0(0,n,wud[0],s0);

      for (i=0;i<=n;i++)
         s0[i]*=(2.0/3.0);

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,-eps,wad,wud[0]);
      action0(0,n,wud[0],s1);

      for (i=0;i<=n;i++)
         s1[i]*=(2.0/3.0);

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,2.0*eps,wad,wud[0]);
      action0(0,n,wud[0],sw);

      for (i=0;i<=n;i++)
         s0[i]-=(sw[i]/12.0);

      copy_ud(n,wsv[0],wud[0]);
      rot_ud(n,-2.0*eps,wad,wud[0]);
      action0(0,n,wud[0],sw);

      for (i=0;i<=n;i++)
         s1[i]-=(sw[i]/12.0);

      for (i=0;i<=n;i++)
         sw[i]=(s0[i]-s1[i])/eps-dS[i];

      if (NPROC>1)
      {
         MPI_Reduce(sw,s0,n+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(s0,n+1,MPI_DOUBLE,0,MPI_COMM_WORLD);

         for (i=0;i<=n;i++)
            sw[i]=s0[i];

         MPI_Reduce(dS,s0,n+1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
         MPI_Bcast(s0,n+1,MPI_DOUBLE,0,MPI_COMM_WORLD);

         for (i=0;i<=n;i++)
            dS[i]=s0[i];
      }

      if (my_rank==0)
      {
         dmax=0.0;

         for (i=0;i<=n;i++)
         {
            d=fabs(sw[i]/dS[i]);

            if (d>dmax)
               dmax=d;
         }

         printf("n = %d, maximal relative deviation = %.1e, ",n,dmax);
         printf("linearity check = %.1e\n",dev);
      }
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
