/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of action0() for plane-wave gauge potentials.
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "utils.h"
#include "su3fcts.h"
#include "random.h"
#include "lattice.h"
#include "pmdb.h"
#include "forces.h"
#include "global.h"

#define N0 (NPROC0*L0)
#define N1 (NPROC1*L1)
#define N2 (NPROC2*L2)
#define N3 (NPROC3*L3)
#define NMAX 4

static const int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
static int opv[(int)(PM_TYPES)];
static double *pv[(int)(PM_TYPES)];
static double mom[NMAX+1][4],fact[NMAX+1];
static double fta[3][NMAX+1],ftu[3][NMAX+1];
static double S0[NMAX+1],S1[NMAX+1],dS[NMAX+1];
static complex_dble Aw[NMAX+1][4];


static void default_pv(int order)
{
   int i,pmax;
   double *p;

   pmax=(int)(PM_TYPES);
   p=malloc(pmax*(order+1)*sizeof(*p));
   error(p==NULL,1,"default_pv [check2.c]",
         "Unable to allocated parameter array");

   for (i=0;i<(pmax*(order+1));i++)
      p[i]=0.0;

   for (i=0;i<pmax;i++)
   {
      opv[i]=0;
      pv[i]=p;
      p+=order+1;
   }

   pv[(int)(C0)][0]=1.0;
   pv[(int)(LAMBDA)][0]=1.0;
   pv[(int)(MASS)][0]=0.0;
   pv[(int)(CG)][0]=1.0;
   pv[(int)(CGP)][0]=1.0;
   pv[(int)(CSW)][0]=1.0;
   pv[(int)(CF)][0]=1.0;
   pv[(int)(CFP)][0]=1.0;

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CG)]=i;
      pv[(int)(CG)][i]=0.126-0.378*(double)(i*i/(order*order));
   }

   for (i=2;i<=order;i+=2)
   {
      opv[(int)(CGP)]=i;
      pv[(int)(CGP)][i]=0.0523+0.117*(double)(i*i/(order*order));
   }

   set_nspt_parms(KSPT,order,opv,pv);
}


static void set_Aw(void)
{
   gauss_dble((double*)(Aw[0]),(NMAX+1)*8);
   MPI_Bcast(Aw[0],(NMAX+1)*8,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void set_mom(void)
{
   int n,bc,mu;
   double pi,r[4],p[4];

   bc=bc_type();
   pi=4.0*atan(1.0);

   r[0]=pi/(double)(N0);
   r[1]=2.0*pi/(double)(N1);
   r[2]=2.0*pi/(double)(N2);
   r[3]=2.0*pi/(double)(N3);

   for (n=0;n<=NMAX;n+=2)
   {
      ranlxd(p,4);

      if (bc==1)
         mom[n][0]=r[0]*floor(p[0]*(double)(N0));
      else
         mom[n][0]=r[0]*(floor(p[0]*(double)(N0))+0.5);

      mom[n][1]=r[1]*floor(p[1]*(double)(N1));
      mom[n][2]=r[2]*floor(p[2]*(double)(N2));
      mom[n][3]=r[3]*floor(p[3]*(double)(N3));

      if (n<NMAX)
      {
         for (mu=0;mu<4;mu++)
            mom[n+1][mu]=-mom[n][mu];
      }
   }

   MPI_Bcast(mom[0],(NMAX+1)*4,MPI_DOUBLE,0,MPI_COMM_WORLD);
}


static void random_flds(int n,su3_alg_dble *ad)
{
   int ix,ifc,i,bc;
   int x0,x1,x2,x3,bo[4],x[4];
   double px0,pxk,A11,A22;
   complex_dble z,w;
   su3_alg_dble *A;

   set_Aw();
   set_mom();

   bc=bc_type();
   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               ix=ipt[x3+L3*x2+L2*L3*x1+L1*L2*L3*x0];

               if (ix>=(VOLUME/2))
               {
                  A=ad+8*(ix-(VOLUME/2))*(n+1);

                  x[0]=x0+bo[0];
                  x[1]=x1+bo[1];
                  x[2]=x2+bo[2];
                  x[3]=x3+bo[3];

                  for (ifc=0;ifc<8;ifc++)
                  {
                     if ((ifc==1)&&(x[0]==0))
                        x[0]=N0-1;
                     else if (ifc&0x1)
                        x[ifc/2]-=1;

                     for (i=0;i<=n;i++)
                     {
                        px0=mom[i][0]*(double)(x[0]);
                        pxk=mom[i][1]*(double)(x[1])+
                           mom[i][2]*(double)(x[2])+
                           mom[i][3]*(double)(x[3]);

                        if (ifc<2)
                        {
                           if (bc==1)
                           {
                              z.re=cos(px0+0.5*mom[i][0])*cos(pxk);
                              z.im=cos(px0+0.5*mom[i][0])*sin(pxk);
                           }
                           else
                           {
                              z.re=-sin(px0+0.5*mom[i][0])*sin(pxk);
                              z.im= sin(px0+0.5*mom[i][0])*cos(pxk);
                           }
                        }
                        else
                        {
                           if (bc==1)
                           {
                              z.re=-sin(px0)*sin(pxk+0.5*mom[i][ifc/2]);
                              z.im= sin(px0)*cos(pxk+0.5*mom[i][ifc/2]);
                           }
                           else
                           {
                              z.re=cos(px0)*cos(pxk+0.5*mom[i][ifc/2]);
                              z.im=cos(px0)*sin(pxk+0.5*mom[i][ifc/2]);
                           }
                        }

                        w=Aw[i][ifc/2];
                        A11=z.re*w.re-z.im*w.im;
                        A22=z.re*w.im+z.im*w.re;
                        su3alg_zero(A);
                        (*A).c1=(1.0/3.0)*(A11-A22);
                        (*A).c2=(1.0/3.0)*(2.0*A11+A22);
                        A+=1;
                     }

                     if ((ifc==1)&&(x[0]==(N0-1)))
                        x[0]=0;
                     else if (ifc&0x1)
                        x[ifc/2]+=1;
                  }
               }
            }
         }
      }
   }
}


static void set_S0(int n,su3_alg_dble *ad)
{
   int bc,m,ix,t,ifc;
   su3_dble **wud,*ud;

   bc=bc_type();
   m=n+1;
   wud=reserve_wud(n+1);
   ud=wud[0];

   for (ix=(VOLUME/2);ix<VOLUME;ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((t>0)||(ifc<2)||(bc!=1))
            pexp_su3alg(m,ad,ud);
         else
            cm3x3_zero(m,ud);

         ad+=m;
         ud+=m;
      }
   }

   action0(1,n,wud[0],S0);
   release_wud();
}


static void set_fact(void)
{
   int i;
   double r;

   r=1.0;

   for (i=0;i<=NMAX;i++)
   {
      r/=(double)(i+2);

      if ((i%4)==2)
         fact[i]=r;
      else if ((i%4)==0)
         fact[i]=-r;
      else
         fact[i]=0.0;
   }
}


static void cos_poly(int n,double *fa,double *fu)
{
   int k,l,j;

   for (k=n;k>=0;k--)
   {
      for (l=(n-k);l>=0;l--)
      {
         fu[l]=fact[k]*fa[l];

         for (j=0;j<l;j++)
            fu[l]+=fu[l-j-1]*fa[j];
      }
   }

   for (l=n;l>=0;l--)
   {
      fu[l]=fa[0]*fu[l];

      for (j=1;j<=l;j++)
         fu[l]+=fu[l-j]*fa[j];
   }
}


static void set_ftu(int n,int pln,int *x)
{
   int mu,nu,bc,i;
   double px0,pxk,ph[2];
   complex_dble z,w;

   mu=plns[pln][0];
   nu=plns[pln][1];
   bc=bc_type();

   for (i=0;i<=n;i++)
   {
      px0=mom[i][0]*(double)(x[0]);
      pxk=mom[i][1]*(double)(x[1])+
         mom[i][2]*(double)(x[2])+
         mom[i][3]*(double)(x[3]);

      if (mu==0)
      {
         if (bc==1)
         {
            z.re=-cos(px0+0.5*mom[i][0])*sin(pxk+0.5*mom[i][nu]);
            z.im= cos(px0+0.5*mom[i][0])*cos(pxk+0.5*mom[i][nu]);
         }
         else
         {
            z.re=-sin(px0+0.5*mom[i][0])*cos(pxk+0.5*mom[i][nu]);
            z.im=-sin(px0+0.5*mom[i][0])*sin(pxk+0.5*mom[i][nu]);
         }
      }
      else
      {
         if (bc==1)
         {
            z.re=-sin(px0)*cos(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
            z.im=-sin(px0)*sin(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
         }
         else
         {
            z.re=-cos(px0)*sin(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
            z.im= cos(px0)*cos(pxk+0.5*mom[i][mu]+0.5*mom[i][nu]);
         }
      }

      ph[0]=2.0*sin(0.5*mom[i][mu]);
      ph[1]=2.0*sin(0.5*mom[i][nu]);

      w.re=ph[0]*Aw[i][nu].re-ph[1]*Aw[i][mu].re;
      w.im=ph[0]*Aw[i][nu].im-ph[1]*Aw[i][mu].im;

      fta[0][i]=z.re*w.re-z.im*w.im;
      fta[1][i]=z.re*w.im+z.im*w.re;
      fta[2][i]=-fta[0][i]-fta[1][i];
   }

   for (i=0;i<3;i++)
      cos_poly(n,fta[i],ftu[i]);
}


static void set_S1(int n)
{
   int bc,pln,is,i,j;
   int x0,x1,x2,x3,bo[4],x[4];
   double *cG,*cGP;

   cG=pv[(int)(CG)];
   cGP=pv[(int)(CGP)];

   bc=bc_type();
   is=init_hsum(n+1);
   reset_hsum(is);

   bo[0]=cpr[0]*L0;
   bo[1]=cpr[1]*L1;
   bo[2]=cpr[2]*L2;
   bo[3]=cpr[3]*L3;

   for (x0=0;x0<L0;x0++)
   {
      for (x1=0;x1<L1;x1++)
      {
         for (x2=0;x2<L2;x2++)
         {
            for (x3=0;x3<L3;x3++)
            {
               x[0]=x0+bo[0];
               x[1]=x1+bo[1];
               x[2]=x2+bo[2];
               x[3]=x3+bo[3];

               for (pln=0;pln<6;pln++)
               {
                  set_ftu(n,pln,x);

                  if ((x[0]>0)||(pln<3)||(bc!=1))
                  {
                     if ((x[0]==0)&&(pln>=3))
                     {
                        for (i=0;i<=n;i++)
                           S1[i]=-1.0*(ftu[0][i]+ftu[1][i]+ftu[2][i]);
                     }
                     else
                     {
                        for (i=0;i<=n;i++)
                           S1[i]=-2.0*(ftu[0][i]+ftu[1][i]+ftu[2][i]);
                     }

                     if (((bc==1)&&(x[0]==0)&&(pln<3))||
                         ((bc==1)&&(x[0]==(N0-1))&&(pln<3))||
                         ((bc==2)&&(x[0]==0)&&(pln>=3)))
                     {
                        for (i=n;i>=0;i--)
                        {
                           for (j=1;(j<=n)&&(i>=j);j+=1)
                              S1[i]+=cG[j]*S1[i-j];
                        }
                     }

                     if ((bc==2)&&(x[0]==(N0-1))&&(pln<3))
                     {
                        for (i=n;i>=0;i--)
                        {
                           for (j=1;(j<=n)&&(i>=j);j+=1)
                              S1[i]+=cGP[j]*S1[i-j];
                        }
                     }

                     add_to_hsum(is,S1);
                  }
               }
            }
         }
      }
   }

   global_hsum(is,S1);

   for (i=0;i<=n;i++)
      dS[i]=S0[i]-S1[i];
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,i;
   su3_alg_dble **wad;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);

      printf("\n");
      printf("Check of action0() for plane-wave gauge potentials\n");
      printf("--------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");
      error_root((bc==0)||(sscanf(argv[bc+1],"%d",&bc)!=1),1,
                 "main [check2.c]","Syntax: check2 -bc <type>");
      fflush(flog);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   default_pv(NMAX);
   print_th_parms();
   print_nspt_parms((1<<(int)(LAMBDA))^0xffff);
   start_ranlux(0,12345);
   geometry();

   set_fact();
   alloc_wad(NMAX+1);
   alloc_wud(NMAX+1);

   for (n=0;n<=NMAX;n++)
   {
      wad=reserve_wad(n+1);
      random_flds(n,wad[0]);
      set_S0(n,wad[0]);
      set_S1(n);

      if (my_rank==0)
      {
         printf("n = %d:\n",n);

         for (i=0;i<=n;i++)
         {
            printf("i = %d, action = % .6e, relative deviation = %.1e\n",
                   i,S0[i],fabs(dS[i]/S0[i]));
         }
      }

      release_wad();
   }

   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
