
/*******************************************************************************
*
* File check2.c
*
* Copyright (C) 2015 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs set_adbc() and check_adbc().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "pmdb.h"
#include "random.h"
#include "utils.h"
#include "aflds.h"
#include "linalg.h"
#include "lattice.h"
#include "global.h"

#define NMAX 4

static void new_flds(int n,su3_alg_dble *ad)
{
   if (cpr[0]==(NPROC0-1))
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4)+3)*(n+1));
   else
      gauss_dble((double*)(ad),8*(4*VOLUME+7*(BNDRY/4))*(n+1));
}


static int is_equal(su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int ie;

   ie=1;
   ie&=((*ad1).c1==(*ad2).c1);
   ie&=((*ad1).c2==(*ad2).c2);
   ie&=((*ad1).c3==(*ad2).c3);
   ie&=((*ad1).c4==(*ad2).c4);
   ie&=((*ad1).c5==(*ad2).c5);
   ie&=((*ad1).c6==(*ad2).c6);
   ie&=((*ad1).c7==(*ad2).c7);
   ie&=((*ad1).c8==(*ad2).c8);

   return ie;
}


static int is_zero(su3_alg_dble *ad)
{
   int ie;

   ie=1;
   ie&=((*ad).c1==0.0);
   ie&=((*ad).c2==0.0);
   ie&=((*ad).c3==0.0);
   ie&=((*ad).c4==0.0);
   ie&=((*ad).c5==0.0);
   ie&=((*ad).c6==0.0);
   ie&=((*ad).c7==0.0);
   ie&=((*ad).c8==0.0);

   return ie;
}


static int cmp_active(int n,su3_alg_dble *ad1,su3_alg_dble *ad2)
{
   int bc,ix,t,ifc,i,m,ie;

   bc=bc_type();
   m=n+1;
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      for (ifc=0;ifc<8;ifc++)
      {
         if ((ifc<2)||(t>0)||(bc!=1))
         {
            for (i=0;i<m;i++)
               ie&=is_equal(ad1+i,ad2+i);
         }

         ad1+=m;
         ad2+=m;
      }
   }

   return ie;
}


static int check_bval(int n,su3_alg_dble *ad)
{
   int bc,ix,t,ifc,i,m,ie;

   bc=bc_type();
   m=n+1;
   ie=1;

   for (ix=(VOLUME/2);ix<(VOLUME);ix++)
   {
      t=global_time(ix);

      if ((bc==1)&&(t==0))
      {
         for (ifc=0;ifc<8;ifc++)
         {
            if (ifc>1)
            {
               for (i=0;i<m;i++)
                  ie&=is_zero(ad+i);
            }

            ad+=m;
         }
      }
      else
         ad+=8*m;
   }

   if (cpr[0]==(NPROC0-1))
   {
      ad+=7*(BNDRY/4)*m;

      for (i=0;i<(3*m);i++)
         ie&=is_zero(ad+i);
   }

   return ie;
}


int main(int argc,char *argv[])
{
   int my_rank,bc,n,ie;
   su3_alg_dble **wad;
   th_parms_t thp;
   FILE *flog=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check2.log","w",stdout);
      printf("\n");
      printf("Check of set_adbc() and check_adbc()\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      bc=find_opt(argc,argv,"-bc");

      if (bc!=0)
         error_root(sscanf(argv[bc+1],"%d",&bc)!=1,1,"main [check2.c]",
                    "Syntax: check2 [-bc <type>]");
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   set_th_parms(LYM,bc,0);
   print_th_parms();
   start_ranlux(0,12345);
   geometry();

   thp=th_parms();
   error(thp.bc!=bc,1,"main [check2.c]",
         "Type of boundary condition is not properly set");
   alloc_wad(2*NMAX);

   for (n=0;n<NMAX;n++)
   {
      wad=reserve_wad(2*(n+1));
      new_flds(n,wad[0]);
      new_flds(n,wad[n+1]);

      assign_alg2alg((4*VOLUME+7*(BNDRY/4))*(n+1),wad[0],wad[n+1]);
      set_adbc(n,wad[0]);

      ie=check_adbc(n,wad[0]);
      error(ie!=1,1,"main [check2.c]",
            "check_adbc() is incorrect");

      ie=check_adbc(n,wad[n+1]);
      error(ie!=0,2,"main [check2.c]",
            "check_adbc() is incorrect");

      ie=cmp_active(n,wad[0],wad[n+1]);
      error(ie!=1,1,"main [check2.c]",
            "Active link variables are modified by set_adbc()");

      ie=check_bval(n,wad[0]);
      error(ie!=1,1,"main [check2.c]",
            "Boundary values are not properly set by set_adbc()");

      release_wad();
   }

   if (my_rank==0)
   {
      printf("No errors detected\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
