
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2011-2013, 2015, 2017 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of set_mdsteps().
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "pmdb.h"
#include "utils.h"
#include "lattice.h"
#include "mdint.h"
#include "global.h"

static int my_rank;


static void read_smd_parms(void)
{
   int nlv,nact,iact[1];
   double eps,gamma,lambda0;

   if (my_rank==0)
   {
      find_section("SMD parameters");
      read_line("nlv","%d",&nlv);
      read_line("eps","%lf",&eps);
      read_line("gamma","%lf",&gamma);
      read_line("lambda0","%lf",&lambda0);
   }

   MPI_Bcast(&nlv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&eps,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&gamma,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&lambda0,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   nact=1;
   iact[0]=0;
   set_smd_parms(nact,iact,0,0,NULL,nlv,eps,gamma,lambda0);
}


static void read_integrator(void)
{
   int nlv,i;
   smd_parms_t smd;

   smd=smd_parms();
   nlv=smd.nlv;

   for (i=0;i<nlv;i++)
      read_mdint_parms(i);
}


int main(int argc,char *argv[])
{
   int i,ie;
   int nop,*iop,*ifr;
   double *eps;
   mdstep_t *mds;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
      fin=freopen("check1.in","r",stdin);

      printf("\n");
      printf("Check of set_mdsteps()\n");
      printf("----------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
   }

   read_smd_parms();
   read_integrator();

   if (my_rank==0)
      fclose(fin);

   set_mdsteps();
   print_smd_parms();
   print_mdint_parms();
   print_mdsteps(0x3);
   mds=mdsteps(&nop);
   i=0;
   ie=(mds[0].top!=TR);

   while (mds[i].top!=TOPS)
      i+=1;

   error(i!=nop,1,"main [check1.c]",
         "Parameter nop returned by mdsteps is incorrect");
   error(ie!=0,1,"main [check1.c]",
         "Trajectory does not start with TR");

   iop=malloc(2*nop*sizeof(*iop));
   eps=malloc(nop*sizeof(*eps));
   error((iop==NULL)||(eps==NULL),1,"main [check1.c]",
         "Unable to allocate auxiliary arrays");
   ifr=iop+nop;

   for (i=0;i<nop;i++)
   {
      iop[i]=(int)(mds[i].top);
      ifr[i]=mds[i].ifr;
      eps[i]=mds[i].eps;
   }

   MPI_Bcast(iop,2*nop,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(eps,nop,MPI_DOUBLE,0,MPI_COMM_WORLD);

   for (i=0;i<nop;i++)
   {
      ie|=(iop[i]!=(int)(mds[i].top));
      ie|=(ifr[i]!=mds[i].ifr);
      ie|=(eps[i]!=mds[i].eps);
   }

   error(ie!=0,1,"main [check1.c]",
         "Integration steps are not globally the same");

   if (my_rank==0)
      fclose(flog);

   MPI_Finalize();
   exit(0);
}
