
/*******************************************************************************
*
* File check5.c
*
* Copyright (C) 2015, 2016 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs sl3_to_zero(),..,su3_mulr_add_sl3().
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "utils.h"
#include "random.h"
#include "su3fcts.h"

#define NTEST 100
#define SEED 376

static su3_alg_dble wa[9] ALIGNED16;
static su3_dble wu[9] ALIGNED16;
static su3_dble Aw[9] ALIGNED16;
static su3_dble Xw[9] ALIGNED16;
static su3_dble Yw[9] ALIGNED16;
static su3_dble Sw[9] ALIGNED16;


static double dev_uv(su3_dble *u,su3_dble *v)
{
   int i;
   double r[18],dev,dmax;

   r[ 0]=(*u).c11.re-(*v).c11.re;
   r[ 1]=(*u).c11.im-(*v).c11.im;
   r[ 2]=(*u).c12.re-(*v).c12.re;
   r[ 3]=(*u).c12.im-(*v).c12.im;
   r[ 4]=(*u).c13.re-(*v).c13.re;
   r[ 5]=(*u).c13.im-(*v).c13.im;

   r[ 6]=(*u).c21.re-(*v).c21.re;
   r[ 7]=(*u).c21.im-(*v).c21.im;
   r[ 8]=(*u).c22.re-(*v).c22.re;
   r[ 9]=(*u).c22.im-(*v).c22.im;
   r[10]=(*u).c23.re-(*v).c23.re;
   r[11]=(*u).c23.im-(*v).c23.im;

   r[12]=(*u).c31.re-(*v).c31.re;
   r[13]=(*u).c31.im-(*v).c31.im;
   r[14]=(*u).c32.re-(*v).c32.re;
   r[15]=(*u).c32.im-(*v).c32.im;
   r[16]=(*u).c33.re-(*v).c33.re;
   r[17]=(*u).c33.im-(*v).c33.im;

   dmax=0.0;

   for (i=0;i<18;i++)
   {
      dev=fabs(r[i]);
      if (dev>dmax)
         dmax=dev;
   }

   return dmax;
}


int main(void)
{
   int n,i,j,k;
   double r,d,dmax;

   printf("\n");
   printf("Check of the programs pexp_su3alg() and mexp_su3alg()\n");
   printf("-----------------------------------------------------\n\n");

#if (defined AVX)
   printf("Using AVX instructions\n");
#elif (defined x64)
   printf("Using SSE3 instructions and up to 16 xmm registers\n");
#endif

   printf("Test performed on %d random matrices\n\n",NTEST);

   rlxd_init(1,SEED);
   dmax=0.0;

   for (i=0;i<NTEST;i++)
   {
      for (n=1;n<=8;n++)
      {
         su3alg_zero(wa);
         cm3x3_unity(1,wu);
         gauss_dble((double*)(wa+1),n*8);
         gauss_dble((double*)(wu+1),n*18);
         r=0.1;

         for (j=1;j<=n;j++)
            su3alg_mulr(&r,wa+j,wa+j);

         pexp_su3alg(n,wa+1,wu+1);

         for (j=0;j<=n;j++)
         {
            su3alg_to_su3(wa+j,Aw+j);
            Xw[j]=Aw[j];
         }

         cm3x3_unity(1,Sw);
         cm3x3_assign(n,Xw+1,Sw+1);
         r=1.0;

         for (j=1;j<n;j++)
         {
            r/=(double)(j+1);
            cm3x3_mul_poly(n,Xw,Aw,Yw);
            cm3x3_assign(n+1,Yw,Xw);

            for (k=1;k<=n;k++)
               cm3x3_mulr_add(&r,Xw+k,Sw+k);
         }

         for (j=0;j<=n;j++)
         {
            d=dev_uv(wu+j,Sw+j);
            if (d>dmax)
               dmax=d;
         }

         cm3x3_log_poly(n,wu+1,Sw+1);

         for (j=1;j<=n;j++)
         {
            d=dev_uv(Sw+j,Aw+j);
            if (d>dmax)
               dmax=d;
         }
      }
   }

   printf("Maximal deviation of pexp_su3alg() = %.1e\n",dmax);
   dmax=0.0;

   for (i=0;i<NTEST;i++)
   {
      for (n=1;n<=8;n++)
      {
         su3alg_zero(wa);
         cm3x3_unity(1,wu);
         gauss_dble((double*)(wa+1),n*8);
         gauss_dble((double*)(wu+1),n*18);
         r=0.1;

         for (j=0;j<=n;j++)
            su3alg_mulr(&r,wa+j,wa+j);

         mexp_su3alg(n,wa+1,wu+1);
         r=-1.0;

         for (j=0;j<=n;j++)
         {
            su3alg_to_su3(wa+j,Aw+j);
            cm3x3_mulr(&r,Aw+j,Aw+j);
            Xw[j]=Aw[j];
         }

         cm3x3_unity(1,Sw);
         cm3x3_assign(n,Xw+1,Sw+1);
         r=1.0;

         for (j=1;j<n;j++)
         {
            r/=(double)(j+1);
            cm3x3_mul_poly(n,Xw,Aw,Yw);
            cm3x3_assign(n+1,Yw,Xw);

            for (k=1;k<=n;k++)
               cm3x3_mulr_add(&r,Xw+k,Sw+k);
         }

         for (j=0;j<=n;j++)
         {
            d=dev_uv(wu+j,Sw+j);
            if (d>dmax)
               dmax=d;
         }

         cm3x3_log_poly(n,wu+1,Sw+1);

         for (j=1;j<=n;j++)
         {
            d=dev_uv(Sw+j,Aw+j);
            if (d>dmax)
               dmax=d;
         }
      }
   }

   printf("Maximal deviation of mexp_su3alg() = %.1e\n\n",dmax);
   exit(0);
}
