
/*******************************************************************************
*
* File pmdb.h
*
* Copyright (C) 2015, 2017 Martin Luescher, Mattia Dalla Brida
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
*******************************************************************************/

#ifndef PMDB_H
#define PMDB_H

#ifndef SU3_H
#include "su3.h"
#endif

/* Named parameters */

typedef enum
{
   ACG,
   ACTIONS
} action_t;

typedef enum
{
   EXP,SIN,COS,
   DFT_TYPES
} dft_type_t;

typedef enum
{
   FRG,
   FORCES
} force_t;

typedef enum
{
   LPFR,OMF2,OMF4,OMF6,
   INTEGRATORS
} integrator_t;

typedef enum
{
   KSPT,LSPT,
   NSPT_TYPES
} nspt_type_t;

typedef enum
{
   C0,LAMBDA,MASS,CSW,CG,CGP,CF,CFP,
   PM_TYPES
} pm_type_t;

typedef enum
{
   PHI4,LYM,LQCD,
   THEORIES
} theory_t;

/* Parameter sets */

typedef struct
{
   dft_type_t type;
   int n,b,c,*r;
   complex_dble *w;
   complex_dble *wb,*wc,*iwb,*iwc;
} dft_parms_t;

typedef struct
{
   int csize;
   int ny[4],nf[4];
   int *nx[4],*mf[4];
   dft_parms_t *dp[4];
} dft4d_parms_t;

typedef struct
{
   int rule;
   double eps,lambda0;
} lang_parms_t;

typedef struct
{
   integrator_t integrator;
   double lambda;
   int nstep,nfr;
   int *ifr;
} mdint_parms_t;

typedef struct
{
   nspt_type_t type;
   int order,opv[(int)(PM_TYPES)];
   double *pv[(int)(PM_TYPES)];
} nspt_parms_t;

typedef struct
{
   int nact,npf,nmu,nlv;
   int *iact;
   double eps,gamma,lambda0;
   double *mu;
} smd_parms_t;

typedef struct
{
   theory_t th;
   int bc,Nf;
} th_parms_t;

/* DFT_PARMS_C */
extern int set_dft_parms(dft_type_t type,int n,int b,int c);
extern dft_parms_t *dft_parms(int id);

/* DFT4D_PARMS_C */
extern int set_dft4d_parms(int *idp,int *nx,int csize);
extern dft4d_parms_t *dft4d_parms(int id);

/* LANG_PARMS_C */
extern lang_parms_t set_lang_parms(int rule,double eps,double lambda0);
extern lang_parms_t lang_parms(void);
extern void print_lang_parms(void);
extern void write_lang_parms(FILE *fdat);
extern void check_lang_parms(FILE *fdat);

/* MDINT_PARMS_C */
extern mdint_parms_t set_mdint_parms(int ilv,
                                     integrator_t integrator,double lambda,
                                     int nstep,int nfr,int *ifr);
extern mdint_parms_t mdint_parms(int ilv);
extern void read_mdint_parms(int ilv);
extern void print_mdint_parms(void);
extern void write_mdint_parms(FILE *fdat);
extern void check_mdint_parms(FILE *fdat);

/* NSPT_PARMS_C */
extern nspt_parms_t set_nspt_parms(nspt_type_t type,int order,
                                   int *opv,double **pv);
extern nspt_parms_t nspt_parms(void);
extern void print_nspt_parms(int mask);
extern void write_nspt_parms(FILE *fdat);
extern void check_nspt_parms(FILE *fdat);

/* SMD_PARMS_C */
extern smd_parms_t set_smd_parms(int nact,int *iact,int npf,int nmu,double *mu,
                                 int nlv,double eps,double gamma,
                                 double lambda0);
extern smd_parms_t smd_parms(void);
extern void print_smd_parms(void);
extern void write_smd_parms(FILE *fdat);
extern void check_smd_parms(FILE *fdat);

/* TH_PARMS_C */
extern th_parms_t set_th_parms(theory_t th,int bc,int Nf);
extern th_parms_t th_parms(void);
extern void print_th_parms(void);
extern void write_th_parms(FILE *fdat);
extern void check_th_parms(FILE *fdat);
extern int bc_type(void);

#endif
